# -*- charset: utf-8 -*-

"""\
An implementation of the authentication features required to work with
mod_auth_tkt from Python.

This is basically a port of all the Perl functionality and examples contained
in mod_auth_tkt.

.. note ::

    This code all looks a bit of a mess from a Python progammers point of 
    view but I've tried hard to keep the implemenation here as close as 
    possible to the original Perl version so that any changes to that 
    version can be easily ported back to this version. As a result the
    code isn't written in the way I would normally write this sort of 
    module but on this occassion I judged maintainability to be more 
    important than style.

"""

############################################################################
#
# Apache Configuration Parsing 
#
# This section implements the Perl ``Apache::AuthTkt`` code 
#

# It would be more sensible if mod_auth_tkt passed the configuration via an
# environment variable but it doesn't so we have to parse the configuration
# manually, which means strucutring a virtual host so that all the AuthTkt
# options are ``include``d from thier own file an parsed like with this parser.

import sys
import re
import logging 
import datetime

log = logging.getLogger('auth_tkt')

# Now from the options, generate
PREFIX = 'TKTAuth'
DEFAULTS = {
    'TKTAuthCookieName':     'auth_tkt',
    'TKTAuthBackArgName':    'back',
    'TKTAuthTimeout':        2 * 60 * 60,
    'TKTAuthTimeoutMin':     2 * 60,
    'TKTAuthTimeoutRefresh': 0.5,
    'TKTAuthGuestLogin':     0,
    'TKTAuthGuestUser':      'guest',
    'TKTAuthIgnoreIP':       0,
    'TKTAuthRequireSSL':     0,
    'TKTAuthCookieSecure':   0,
}

def parse_conf(filename, DEFAULTS=DEFAULTS, PREFIX=PREFIX, ignore_secret=False):
    config = {}
    config_file = open(filename, 'r')
    for line in config_file:
        # Get rid of \n
        line = line.strip()
        # Empty?
        if not line:
            continue
        if line.startswith("#") or line.startswith('<'):
            continue
        parts = line.split(" ")
        name = parts[0]
        value = ' '.join(parts[1:])
        # Strip leading " characters.
        if (value[0] == value[-1] == '"') and len(value)>1:
            value = value[1:-1]
        name = name.strip()
        if name.startswith(PREFIX) and not config.has_key('name'):
            log.debug("Found %r %r", name, value)
            config[name] = value
    config_file.close() 

    # Merge in defaults
    d = DEFAULTS.copy()
    d.update(config)
    config = d

    if not config.has_key('TKTAuthSecret') and not ignore_secret:
        raise Exception(
            'TKTAuthSecret directive not found in config file %r'%filename
        )

    # The perl implementation obtains this from the environment HTTP_HOST or SERVER_NAME
    # but to me that defeats the object of having it surely?
    #if not config.has_key('TKTAuthDomain'):
    #    raise Exception('TKTAuthDomain directive not found')

    final = {}
    config_map = {
        'secret':            'Secret',
        'cookie_name':       'CookieName',
        'back_cookie_name':  'BackCookieName',
        'back_arg_name':     'BackArgName',
        'domain':            'Domain',
        'cookie_expires':    'CookieExpires',
        'login_url':         'LoginURL',
        'timeout_url':       'TimeoutURL',
        'post_timeout_url':  'PostTimeoutURL',
        'unauth_url':        'UnauthURL',
        'timeout':           'Timeout',
        'timeout_min':       'TimeoutMin',
        'timeout_refresh':   'TimeoutRefresh',
        'token':             'Token',
        'debug':             'Debug',
        'guest_login':       'GuestLogin',
        'guest_user':        'GuestUser',
        'ignore_ip':         'IgnoreIP',
        'require_ssl':       'RequireSSL',
        'cookie_secure':     'CookieSecure',
    }
    convert = {}
    for k, v in config_map.items():
        convert[v] =k 
    for k, v in config.items():
        name = k[len(PREFIX):]
        if k in ['TKTAuthGuestLogin', 'TKTAuthIgnoreIP', 'TKTAuthRequireSSL', 'TKTAuthCookieSecure']:
            value = {'0': 0, '1': 1, 'on':1, 'yes':1, 'true':1, 'off':0, 'no':0, 'false':0}[str(v).lower()]
        elif k in ['TKTAuthCookieExpires', 'TKTAuthTimeout']:
            value = convert_time_seconds(v)
        elif name in config_map.values():
            value = v
        else:
            raise Exception('Unknown directive %r (%r)'%(k, name))


        # No idea why we can't allow the AuthTktDebug setting in the final dictionary but that is
        # what happens in the original.
        if convert[name] != 'debug':
            final[convert[name]] = value
    final['conf'] = filename
    final['debug'] = 0
    return final

def convert_time_seconds(time):
    # Helper routine to convert time units into seconds
    units = { 
      's' : 1,
      'm' : 60,
      'h' : 3600,
      'd' : 86400,
      'w' : 7 * 86400,
      'M' : 30 * 86400,
      'y' : 365 * 86400,
    }
    try:
        return int(time)
    except (ValueError, TypeError), e:
        sec = 0
        mode = 'number'
        amt = ''
        time = time.replace(' ','')
        for i in range(len(time)):
            if time[i] in units.keys():
                if not amt:
                    raise Exception('Expected a number before %r in %r'%(time[i], time))
                else:
                    sec += int(amt) * units[time[i]]
                    amt = ''
            else:
                amt += time[i]
        if amt:
            sec += int(amt) * units[time[i]]
        return datetime.timedelta(seconds=sec)

# We are re-using the contributes AuthTicket rather than creating our own implemenation.
# Start contributed code ----->

##########################################################################
#
# Copyright (c) 2005 Imaginary Landscape LLC and Contributors.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
##########################################################################

import time as time_mod
import md5
import cgi
import Cookie

class AuthTicket(object):

    """
    This class represents an authentication token.  You must pass in
    the shared secret, the userid, and the IP address.  Optionally you
    can include tokens (a list of strings, representing role names),
    'user_data', which is arbitrary data available for your own use in
    later scripts.  Lastly, you can override the cookie name and
    timestamp.

    Once you provide all the arguments, use .cookie_value() to
    generate the appropriate authentication ticket.  .cookie()
    generates a Cookie object, the str() of which is the complete
    cookie header to be sent.

    CGI usage::

        token = auth_ticket.AuthTicket('sharedsecret', 'username',
            os.environ['REMOTE_ADDR'], tokens=['admin'])
        print 'Status: 200 OK'
        print 'Content-type: text/html'
        print token.cookie()
        print
        ... redirect HTML ...

    Webware usage::

        token = auth_ticket.AuthTicket('sharedsecret', 'username',
            self.request().environ()['REMOTE_ADDR'], tokens=['admin'])
        self.response().setCookie('auth_tkt', token.cookie_value())

    Be careful not to do an HTTP redirect after login; use meta
    refresh or Javascript -- some browsers have bugs where cookies
    aren't saved when set on a redirect.
    """

    def __init__(self, secret, userid, ip, tokens=(), user_data='',
                 time=None, cookie_name='auth_tkt',
                 secure=False):
        self.secret = secret
        self.userid = userid
        self.ip = ip
        self.tokens = ','.join(tokens)
        self.user_data = user_data
        if time is None:
            self.time = time_mod.time()
        else:
            self.time = time
        self.cookie_name = cookie_name
        self.secure = secure

    def ipts(self):
        ip_chars = ''.join(map(chr, map(int, self.ip.split('.'))))
        t = int(self.time)
        ts = ((t & 0xff000000) >> 24,
              (t & 0xff0000) >> 16,
              (t & 0xff00) >> 8,
              t & 0xff)
        ts_chars = ''.join(map(chr, ts))
        return ip_chars + ts_chars

    def digest0(self):
        return md5.new(
            self.ipts() + str(self.secret)
            + str(self.userid) + '\0' + str(self.tokens)
            + '\0' + str(self.user_data)).hexdigest()

    def digest(self):
        return md5.new(self.digest0() + self.secret).hexdigest()

    def cookie_value(self):
        v = '%s%08x%s!' % (self.digest(), int(self.time), self.userid)
        if self.tokens:
            v += self.tokens + '!'
        v += self.user_data
        return v

    def cookie(self):
        c = Cookie.SimpleCookie()
        c[self.cookie_name] = self.cookie_value().encode('base64').strip()
        c[self.cookie_name]['path'] = '/'
        if self.secure:
            c[self.cookie_name]['secure'] = 'true'
        return c

# <----- End contributed code 

# Now we need a class to store the variables and wrap the ticket parsing

class AuthTkt(object):
    """\

    Parse an Apache config file conating the mod_auth_tkt options and provide
    a dictionary-like interface to read them.

    Also provide a ``ticket()`` method to generate a ticket based on the 
    options.

    .. note ::

        This class is designed to be a close approximation to the functionality
        contained in the Perl ``AuthTkt.pm`` module at
        http://search.cpan.org/~gavinc/Apache-AuthTkt-0.08/AuthTkt.pm and is the
        main interface which should be used in an implementation to access the
        AuthTkt* options and generate tickets.

    Used it like this::

        # Constructor - either (preferred):
        at = AuthTkt(conf='/etc/httpd/conf.d/auth_tkt.conf')
        # OR:
        at = AuthTkt(secret='818f9c9d-91ed-4b74-9f48-ff99cfe00a0e')

        # Generate ticket
        ticket = at.ticket(userid=username, ip=ip_addr)
    
        # Get the ticket cookie value:
        value = ticket.cookie_value()

        # Or generate cookie containing ticket
        cookie = ticket.cookie()
    
        # Access the shared secret
        secret = at['secret']
        # If using the 'conf' constructor above, all other TKTAuth attributes 
        #   are also available e.g.:
        at['cookie_name'], at['ignore_ip'], at['request_ssl']

    """
    def __init__(self, conf=None, secret=None):
        if conf is None and secret is None:
            raise Exception('You must specify a config file or secret')
        if conf is not None:
            # This adds the filename itself as the ``conf`` key as well as
            # parsing all the variables.
            self.options = parse_conf(conf, DEFAULTS=DEFAULTS, PREFIX=PREFIX, ignore_secret = secret is not None and True)
        if secret is not None:
            self.options['secret'] = secret

    def __getitem__(self, name):
        if self.options.has_key(name):
            return self.options[name]
        else:
            raise KeyError('No such option %r'%name)

    def get(self, name, default=None):
        if self.options.has_key(name):
            return self.options[name]
        else:
            return default

    def has_key(self, name):
        return self.options.has_key(name)

    def ticket(self, userid, ip, tokens=(), user_data='', time=None):
        if self['ignore_ip']:
            ip = '0.0.0.0'
        return AuthTicket(
            secret=self['secret'],
            userid=userid,
            ip = ip,
            tokens = tokens,
            user_data = user_data,
            time=time,
            cookie_name = self['cookie_name'],
            secure=self['cookie_secure'],
        )
        
    def cookie(self):
        raise NotImplementedError('Just call ticket().cookie() instead')

############################################################################
#
# Default CGI implementation configuration
#
# This section implements the Perl ``AuthTktConfig.pm`` code but is only
# used by CGI examples as a default configuration to pass to the handler
# defined later. It has no bearing on the AuthTkt* options handled by the
# AuthTkt class above, it just affects the defaults the CGI scripts pass
# to the handler.
#

# Note that this is different from the Perl version in that it has a 
# flow argument.
def validate_sub(username, password, flow=None):
    # The default validation always returns False. You should create
    # your own handler to handle authentication. 
    return False

# Deprecated but still in the Perl implementation
def guest_sub():
    return 'guest'

# Not in the Perl implementation but allows customisation
def render_html(q):
    if q.has_key('errors'):
        q['error_str'] = '<p class="error">%s</p>' % ('<br />'.join(q['errors']))
    else:
        q['error_str'] = ''
    return '''
<html>
<head>
    <title>%(title)s</title>
    <script type="text/javascript">
        function getFocus() {
          document.forms[0].elements[0].focus();
          document.forms[0].elements[0].select();
        };
    </script>
</head>
<body onload="getFocus()">
    <h1>%(title)s</h1>
    <p class="warning">Authorized Use Only</p>
    %(error_str)s
    %(debug_vars)s
    <form name="login" method="post" action="%(base)s">
    <table border="0" cellpadding="5">
    <tr><th>Username:</th><td><input type="text" name="username" /></td></tr>
    <tr><th>Password:</th><td><input type="password" name="password" /></td></tr>
    <tr><td></td><td>
    <input type="submit" value="Login" />
    %(back_html_form)s
    </td></tr>
    </table></form>
    %(back_html)s
    </body>
    </html>
    '''%q
    
    
def get_default_handler_configuration():
    return {
        # @@@ Perl implementation seems to have forgotten to implement 
        # nuke cookies so we won't either
        #'nuke_cookies':[],
        'fixed_back_location':None,
        # Implementation fail:
        'default_back_location':None,
        'back_referer': True,
        'autologin_fallback_mode':'login',
        'validate_sub': validate_sub,
        'guest_sub': guest_sub,
        'stylesheet':'tkt.css',
        'title':'Sign In',

        # New options in the Python implementation
        'signed_out_page':'Signed out',
        'signed_in_page':'Signed in',
        'render_signin_page': render_html,
    }

############################################################################
#
# Handler for the login.cgi, logout.cgi and autologin.cgi scripts
#
# The Perl implementation implements this mainly in the ``login.cgi`` script
# but we are doing it as a handler so that the same functionality can be 
# used in WSGI applications.
#

import cgi
import urllib
import urlparse
from httpkit.helper.cookie import get_cookie, set_cookie

def handle(at, conf, base, mode, flow):
                     
    environ = flow.http.environ

    if mode == 'logout':
        response.delete_cookie('auth_tkt', domain=at.get('domain') or server_name, path='/')
        response.write(conf['signed_out_page'])
        return 
        # The perl implementation does this but I don't see why we need it:
        # token = AuthTicket('672h158d-37b8-4244-aec7-f1e1abf882bb', 'username', environ['REMOTE_ADDR'], tokens=['admin'])
        # response.set_cookie('auth_tkt', token.cookie_value(), expires='-1h')

    # @@@ Does at need updating from $ENV{MOD_AUTH_TKT_CONF}?
    # Well it doesn't seem to be present so I guess not.

    # Get the server_name and server_port trying the following in order:
    # HTTP_HOST, SERVER_NAME, SERVER_PORT
    http_host = environ.get('HTTP_HOST')
    if http_host is not None:
        if ':' in http_host:
            server_name, server_port = http_host.split(':')
        else:
            server_name = http_host
            server_port = environ.get('SERVER_PORT')
    else:
        server_name = environ.get('SERVER_NAME')
        server_port = environ.get('SERVER_PORT')

    # Get the auth_domain looking at the 'domain' key in the configuration then
    # the server_name variable defined above

    auth_domain = at.get('domain') or server_name

    # Try to get the cookies ticket (from the name specified in the config as
    # 'cookie_name'), probe (from 'auth_probe') and if the config file has a
    # 'back_cookie_name' option, get that value too.

    cookie = get_cookie(environ.get('HTTP_COOKIE', ''))
    ticket = cookie and cookie.has_key(at['cookie_name']) and cookie[at['cookie_name']].value or None
    probe = cookie and cookie.has_key('auth_probe') and cookie['auth_probe'].value or None


    back_cookie = None
    if at.has_key('back_cookie_name'):
        back_cookie = cookie.has_key(at['back_cookie_name']) and cookie[at['back_cookie_name']].value or None
    have_cookies = ticket or probe or back_cookie or ''

    # Get the back location as a CGI and HTML escaped string, adding the domain
    # if needed.

    back = conf.get('fixed_back_location') or back_cookie
    if not back and at.get('back_arg_name'):
        back = flow.http.input and flow.http.input.has_key(at['back_arg_name']) and flow.http.input[at['back_arg_name']].value or None
    if not back and conf.get('default_back_location'):
        back = conf['default_back_location']
    if not back and conf.get('back_referer') and environ.get('HTTP_REFERER'):
        back =  environ['HTTP_REFERER']
    
    if back:
        if back.startswith('/'):
            # Add in the rest of the URL
            if server_port == 80:
                back = 'http://'+server_name+back
            elif server_port == 443:
                back = 'http://'+server_name+back
            else:
                back = 'http://'+server_name+':'+server_port+back

        if back.startswith('http%3A%2F%2F') or back.startswith('https%3A%2F%2F'):
            back = urllib.unquote(back)
        back_esc = urllib.quote(back)
        back_html = cgi.escape(back)
    else:
        back_html = None
        back_esc = None

    if mode not in ['login', 'guest', 'autologin']:
        # @@@ The perl version sets the mode to login if it isn't sure, we'll be a little stricter
        raise Exception('Invalid mode %r'%mode)

    self_redirect = int(flow.http.query and flow.http.query.has_key('redirect') and flow.http.query['redirect'].value or 0)
    username = flow.http.input and flow.http.input.has_key('username') and flow.http.input['username'].value.lower() or None
    password = flow.http.input and flow.http.input.has_key('password') and flow.http.input['password'].value or None
    timeout = flow.http.input and flow.http.input.has_key('timeout') and flow.http.input['timeout'].value or None
    unauth = flow.http.input and flow.http.input.has_key('unauth') and flow.http.input['unauth'].value or None
    ip_addr = None
    if not at['ignore_ip']:
        ip_addr = environ['REMOTE_ADDR']
    redirected = 0

    # Summarise, we haven't set $location, $suffix

    fatal = None
    errors = []

    # ------------------------------------------------------------------------
    # Set the auth cookie and redirect to $back

    def _set_cookie(tkt, auth_domain, at):
        cookie = tkt.cookie()
        # @@@ I think this should be cookie_secure, but again, I'm following
        # the perl implementation.
        if at['require_ssl']:
            cookie[at['cookie_name']]['secure'] = True
        if at.has_key('cookie_expires'):
            cookie[at['cookie_name']]['expires'] = at['cookie_expires']
        cookie[at['cookie_name']]['domain'] = auth_domain
        parts = str(cookie).split(':')
        flow.http.response.headers.append((parts[0], ':'.join(parts[1:])))
    
    def set_cookie_redirect(tkt, back):
        if not back:
            raise Exception('This code is never called, perhaps ported incorrectly from perl?')
            # Instead, the redirected back is treated as back
            _set_cookie(tkt, auth_domain, at)
            flow.http.response.body = ['<html><head></head><body><p>Login successful</p></body></html>']
            return 0
        
        # Set (local) cookie, and redirect to $back
        _set_cookie(tkt, auth_domain, at)
        if at.get('debug', False):
            return 0
    
        # If back domain doesn't match auth_domain, pass the ticket via back GET param:
        domain = auth_domain or server_name
        b = urlparse.urlparse(back)
        if b.hostname != domain:
            if b.query:
                back += '&'
            else:
                back += '?'
            back += at['cookie_name'] + '=' + tkt.cookie_value()
    
        # Using a Location: header doesn't always set the cookie so we use a meta refresh.
        flow.http.response.body = ["""
            <html><head><title>Redirecting...</title><meta http-equiv="REFRESH" content="0;URL=%s"></head><body></body></html>
            """%back
        ]
        return 1
    
    # ------------------------------------------------------------------------
    # Actual processing
    
    # If no cookies found, first check whether cookies are supported
    if not have_cookies:
        # If this is a self redirect warn the user about cookie support
        if self_redirect:
            fatal = "Your browser does not appear to support cookies or has cookie support disabled.<br />\nThis site requires cookies - please turn cookie support on or try again using a different browser.";
      
        else: 
            # If no cookies and not a redirect, redirect to self to test cookies
            extra = ''
            if timeout:
                extra += 'timeout=1'
            if unauth:
                extra += 'unauth=1'
            if extra:
                extra = '&extra'
            flow.http.response.headers.append(
                (
                    'Set-Cookie', 
                    set_cookie('auth_probe', 1, domain=auth_domain)
                )
            )
            to = base+"?redirect=%s&%s=%s%s"%(
                self_redirect + 1, 
                at['back_arg_name'],
                back_esc or '', 
                extra
            )
            flow.http.response.body = ["""
                <html><head><title>Redirecting...</title><meta http-equiv="REFRESH" content="0;URL=%s"></head><body></body></html>
                """%to
            ]
            redirected = 1;
    if mode == 'login' and ticket:
        flow.http.response.body = [conf['signed_in_page']]
        return 
    elif mode == 'autologin':
        # If we have a ticket, redirect to $back, including ticket as GET param
        if ticket and back and not timeout:
            b = urlparse.urlparse(back)
            if b.query:
                back += '&'
            else:
                back += '?'
            back += at['cookie_name'] + '=' + ticket
            redirected = 1;
            flow.http.response.headers['Location'] = back
            flow.http.response.status = '307 Redirect'
            flow.http.response.body = ['Redirecting...']
            return
        # Can't autologin - change mode to either guest or login
        else:
            mode = conf['autologin_fallback_mode']
    
    if not (fatal or redirected):
        if not at:
            fatal = "AuthTkt error: No config when a fatal or redirected variable was set"
        elif mode == 'login':
            if username:
                if flow.ext.has_key('username'):
                    raise Exception('flow.ext.username already exists')
                if flow.ext.has_key('password'):
                    raise Exception('flow.ext.password already exists')
                flow.ext['username'] = username
                flow.ext['password'] = password
                args = conf['validate_sub'](flow)
                if args['valid'] is True:
                    tkt = at.ticket(
                        userid=username, 
                        ip=ip_addr,
                        tokens=args.get('tokens', []),
                        user_data=args.get('data', ''),
                    )
                    if not errors:
                        redirected = set_cookie_redirect(tkt, back)
                        fatal = "Login successful."
                else:
                    errors.append("Invalid username or password.")
        elif mode == 'guest': 
            # Generate a guest ticket and redirect to $back
            tkt = at.ticket(userid=conf['guest_sub'](), ip=ip_addr)
            if not errors:
                redirected = set_cookie_redirect(tkt, back)
                fatal = "No back link found."
    
    # Render the page
    
    q = {}
    q['style'] = conf.get('stylesheet', '')
    q['title'] = conf.get('title', mode+" Page")
    q['base'] = base
    q['debug_vars'] = ''
    q['back_html_form'] = ''
    q['back_html'] = ''
    if not redirected:
        # If here, either some kind of error or a login page
        if not fatal:
            if timeout:
                errors.append("Your session has timed out.")
            if unauth:
                errors.append("You are not authorised to access this area.")
        if at['debug']:
            cookie_name = at.get('cookie_name')
            back_cookie_name = at.get('back_cookie_name')
            back_arg_name = at.get('back_arg_name')
            cookie_expires = at.get('cookie_expires', 0)
            q['debug_vars'] = """
    <pre>
    server_name: %(server_name)s
    server_port: %(server_port)s
    domain: %(auth_domain)s 
    mode: %(mode)s
    base: %(base)s
    cookie_name: %(cookie_name)s
    cookie_expires: %(cookie_expires)s
    back_cookie_name: %(back_cookie_name)s
    back_arg_name: %(back_arg_name)s
    back: %(back)s
    back_esc: %(back_esc)s
    back_html: %(back_html)s
    have_cookies: %(have_cookies)s
    ip_addr: %(ip_addr)s
    </pre>"""%locals()
        if fatal:
            q['errors'] = [fatal]
        else:
            if errors:
                q['errors'] = errors
            else:
                q['errors'] = []
            if back_html:
                q['back_html_form'] = """<input type="hidden" name="back" value="%s" />\n"""%back_html
        if back_html:
            q['back_html'] = """<p><a href="%s">Previous Page</a></p>\n"""%back_html

        flow.http.response.body = [conf['render_signin_page'](q)]
    return

#
#
# Utiltiy method used by the CGI scripts. Taken from http://www.python.org/dev/peps/pep-0333/
#

import os, sys

def run_with_cgi(application):

    environ = dict(os.environ.items())
    environ['wsgi.input']        = sys.stdin
    environ['wsgi.errors']       = sys.stderr
    environ['wsgi.version']      = (1,0)
    environ['wsgi.multithread']  = False
    environ['wsgi.multiprocess'] = True
    environ['wsgi.run_once']    = True

    if environ.get('HTTPS','off') in ('on','1'):
        environ['wsgi.url_scheme'] = 'https'
    else:
        environ['wsgi.url_scheme'] = 'http'

    headers_set = []
    headers_sent = []

    def write(data):
        if not headers_set:
             raise AssertionError("write() before start_response()")

        elif not headers_sent:
             # Before the first output, send the stored headers
             status, response_headers = headers_sent[:] = headers_set
             sys.stdout.write('Status: %s\r\n' % status)
             for header in response_headers:
                 sys.stdout.write('%s: %s\r\n' % header)
             sys.stdout.write('\r\n')

        sys.stdout.write(data)
        sys.stdout.flush()

    def start_response(status,response_headers,exc_info=None):
        if exc_info:
            try:
                if headers_sent:
                    # Re-raise original exception if headers sent
                    raise exc_info[0], exc_info[1], exc_info[2]
            finally:
                exc_info = None     # avoid dangling circular ref
        elif headers_set:
            raise AssertionError("Headers already set!")

        headers_set[:] = [status,response_headers]
        return write

    result = application(environ, start_response)
    try:
        for data in result:
            if data:    # don't send headers until body appears
                write(data)
        if not headers_sent:
            write('')   # send headers now if body was empty
    finally:
        if hasattr(result,'close'):
            result.close()

